# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_magic_cell.ipynb.

# %% auto 0
__all__ = ['env', 'config', 'chat', 'cells_to_messages', 'CellChat', 'NBClassicChat', 'NoJSChat', 'fr_cell']

# %% ../nbs/01_magic_cell.ipynb 6
system_prompt="""
You are Claude, a very knowledgeable and intelligent research assistant.
Fulfill simple requests right away.
For more complicated tasks, use <thought> to plan your actions.

When appropriate, use Jupyter notebook. It has python 3.10+ and many packages installed.
Use <code> to add a new code cell at the end of your message. You will receive all its output on the next turn.

Prefer short cells that can be tested quickly.

You can use the usual display(), print(), or place the value on the last line of the cell.
Make sure to not add anything after that value, as it won't be returned if it's not on the last line.
Only output a result once with one method.
Avid using .show() and close matplot figures before the end of the cell - use an explicit display().

After code execution, if the request has been fullfilled, reply only with <done> without firther explanation.
The user will ask follow-up questions if needed. Only do only what they asked for.
You should also end your text with <done> if there is no need to run code.

To show an example in python without running the code, use
```python
python code here
```

For example:
User: Calculate square root of pi
Assistant: <code>
import math
math.sqrt(math.pi)
</code>
User: <stdio>1.7724538509055159<stdio>
Assistant: <done>

User: Display cat.jpg and describe wat you see.
Assistant: <thought>I can use the PIL library for this.</thought><code>
from PIL import Image
Image.open("cat.jpg")
</code>
User: <PIL.PngImagePlugin.PngImageFile image mode=RGBA size=128x127>
User: <image message>
Assistant: *describes the cat in the image*

When writing code, use advanced python features. Assume all packages are installed.
Import modules before using them.
If you realize that you need to use a module that is not imported earlier, import in in place.
"""


# %% ../nbs/01_magic_cell.ipynb 7
import sys
import base64
import json
from typing import List, Dict, Tuple
import textwrap
import warnings

from friendlly.notebook import (
    update_code_self,
    update_code_next,
    detect_environment,
    vscode_get_cells,
    nbclassic_patch_kernel,
    nbclassic_add_cell,
    nbclassic_update_cell,
    nbclassic_render_cell,
    nbclassic_execute_cell,
    nbclassic_delete_cell,
    nbclassic_get_cells,
    )


from .utils import nict, parse_cell_line, cell_to_blocks, make_cell_from_captured, merge_messages
from .config import handle_config
from .capture_trace import SaveTraceback
from .capture_display import PassthroughCapturer
from .capture_io import TeeIO
import claudette
from claudette import text_msg, img_msg, mk_msg


from IPython import get_ipython
from IPython.display import display, clear_output, Markdown, Javascript

# %% ../nbs/01_magic_cell.ipynb 8
def cells_to_messages(cells):
    messages = []


    # We found the first %%fr message (request from user).
    # This means that we treat further cells (markdown and code without %%fr) as assistant-generated.
    user_message_found = False
    for cell in cells:
        cell_friendlly, cell_reply, code_text, output_blks = cell_to_blocks(cell)

        if code_text:
            # If the cell is marked with %%fr reply, or we found the user request above.
            if cell_reply:       role="assistant"  # Messages has '%%fr reply'
            elif cell_friendlly: role="user" # Message has %%fr without 'reply'
            else:                role="assistant" if user_message_found else "user" # It's not an %%fr cell.

            # We found an %%fr cell that is not a reply - means it's a user message!
            if cell_friendlly and not cell_reply: user_message_found=True

            # Assistant code gets wrapped into tags, user code gets wrapped into ```
            # Markdown is passed as is.
            if role == "assistant" and cell.cell_type == "code":
                code_text = f"<code>{code_text}</code>"
            elif role == "user":
                code_text = f"```\n{code_text}\n```"

            messages.append(mk_msg([code_text], role=role))

        # Outputs for both user and assistant code cells are always user messages.
        if output_blks:
            messages.append(mk_msg(output_blks, role="user"))

    return [nict(m) for m in messages]

# %% ../nbs/01_magic_cell.ipynb 9
class CellChat():
    def __init__(self, config):
        self.config = config
        # self.display_handle = None # Used when config.md_cell=False
        self.history = []
        self.client = claudette.Client(model=self.config.model)

    def update_markdown(self, text):
        raise NotImplementedError

    def delete_markdown(self):
        raise NotImplementedError

    def update_code(self, code):
        raise NotImplementedError

    def update_code_final(self, code):
        raise NotImplementedError

    def go(self, idx,  num_above, messages):
        self.idx = idx
        self.added_md_cell = self.added_code_cell = False
        self.thought_display_id = None
        self.history.append(messages)

        tokens = []
        try:
            md_reply = ""
            md_updated = False
            code_reply = ""
            code_updated = False
            thought = ""

            state = "md"

            r = self.client(messages,
                            sp=self.config.system_prompt,
                            stream=True,
                            stop=["</code>", "<done>"])
            for token in r:
                tokens.append(token)
#                 print(state, f"'{token}'", f"'{code_reply}'", f"'{md_reply}'")

                if state == "md":
                    md_reply += token
                    md_updated = True

                    if "<code>" in md_reply:
                        md_reply, code_reply = md_reply.split("<code>")
                        code_updated = True
                        state = "code"

                    if "<thought>" in md_reply:
                        md_reply_body, thought = md_reply.split("<thought>")
                        state = "thought"

                elif state == "thought":
                    thought += token
                    md_tail = ""
                    md_updated = True

                    def combine_md_reply(body, thought, tail=""):
                        return f'{body}<details><summary style="display: list-item">🧠</summary>{thought}</details>{tail}'

                    if "</thought>" in thought:
                        thought, md_tail = thought.split("</thought>")
                        state = "md"
                    md_reply = combine_md_reply(md_reply_body, thought, md_tail)

                else: #code
                    code_reply += token
                    code_updated = True

                if md_updated:
                    md_updated = False
                    if md_reply.strip():
                        self.update_markdown(md_reply.strip())
                    else:
                        self.delete_markdown()

                if code_updated:
                    code_updated = False

                    self.update_code(code_reply.strip())

            if code_reply.strip():
                prefix = f"%%fr +{num_above + 2 if self.added_md_cell else 1} reply\n"
                self.update_code_final(prefix + code_reply.strip())

            if self.client.stop_reason == "stop_sequence":
                tokens += self.client.stop_sequence
            self.history.append(claudette.mk_msg("".join(tokens), role="assistant"))

            code_cell_idx = None if self.idx is None else self.idx + int(self.added_code_cell) + int(self.added_md_cell)
            if md_reply and not self.added_md_cell: md_display = Markdown(md_reply)
            else: md_display = None
            # We return md_display to be re-displayed after cleanup if we did not use an extra cell for md.
            return code_cell_idx, md_display

        except BaseException as e:
            raise e
#             display(Markdown(f"🚫 {repr(e)}"))
#             get_ipython().showtraceback()

# %% ../nbs/01_magic_cell.ipynb 10
class NBClassicChat(CellChat):
    def __init__(self, config):
        super().__init__(config)
        self.display_handle = None

    def update_markdown(self, text):
        # md_cells means we output the text as markdown cells.
        # Otherwise, we display them.
        if self.config.md_cells and self.idx is not None:
            if not self.added_md_cell:
                nbclassic_add_cell(self.idx+1, "markdown")
                self.added_md_cell = True
            nbclassic_update_cell(self.idx+1, text)
            nbclassic_render_cell(self.idx+1)
        else:
            if self.display_handle:
                self.display_handle.update(Markdown(text))
            else:
                self.display_handle = display(Markdown(text), display_id=True)

    def delete_markdown(self):
        if self.config.md_cells and self.added_md_cell:
            nbclassic_delete_cell(self.idx+1)
            self.added_md_cell = False
        elif self.display_handle:
            # Empty JS cells ge cleaned up after the cell is executed.
            self.display_handle.update(Javascript(""))

    def update_code(self, code):
        # Without idx, we will have to update the code in the final call using the payload method.
        if self.idx is not None:
            code_idx = self.idx+2 if self.added_md_cell else self.idx+1
            if not self.added_code_cell:
                nbclassic_add_cell(code_idx, "code")
                self.added_code_cell = True

            nbclassic_update_cell(code_idx, code)

    def update_code_final(self, code):
        if self.idx is not None:
            code_idx = self.idx+2 if self.added_md_cell else self.idx+1
            if not self.added_code_cell:
                nbclassic_add_cell(code_idx, "code")
                self.added_code_cell = True

            nbclassic_update_cell(code_idx, code)
        else:
            # We got no index will have to resort to the native ipython method.
            update_code_self(code)

# %% ../nbs/01_magic_cell.ipynb 11
class NoJSChat(CellChat):
    """
    This chat variant does not use JS inhection that is only available in Jupyter nbclassic.
    """

    def __init__(self, config):
        super().__init__(config)
        self.md_handle = None
        self.code_handle = None

    def update_markdown(self, text):
        if self.md_handle:
            self.md_handle.update(Markdown(text))
        else:
            self.md_handle = display(Markdown(text), display_id=True)

    def delete_markdown(self):
        if self.md_handle:
            self.md_handle.update(Javascript(""))

    def update_code(self, code):
        code = Markdown(f"```\n{code}\n```")
        if not self.code_handle:
            self.code_handle = display(Markdown("⌛"), display_id=True)
            self.num_code_updates = 1
        else:
            self.code_handle.update(Markdown(f"{'⏳' if self.num_code_updates % 2 else '⌛'}"))
            self.num_code_updates += 1

    def update_code_final(self, code):
        if self.code_handle:
            self.code_handle.update(Javascript("")) # Will be cleaned up shortly.
        update_code_next(code)

# %% ../nbs/01_magic_cell.ipynb 12
env = detect_environment()

config = nict(
    environment=env,
    autorun=(env == "nbclassic"),
    md_cells=(env == "nbclassic"),
    comment_after_run=True,
    api_key=None,
    model="claude-3-5-sonnet-20240620",
    system_prompt=system_prompt
)

chat = None
def fr_cell(line=None, cell=None):
    global chat
    global config

    if config.environment == "nbclassic":
        chat = NBClassicChat(config=config)
    else:
        chat = NoJSChat(config=config)

    num_cells, is_friendlly, is_reply, is_config = 0, False, False, False
    idx = None
    ip = get_ipython()
    prev_messages = []
    if line:
        # line comes with magic already stripped.
        num_cells, is_friendlly, is_reply, is_config = parse_cell_line("%%fr " + line)
        if num_cells and not is_config:
            assert is_friendlly, "Called on a cell without %%fr?"

            if config.environment == "nbclassic":
                idx, cells_above = nbclassic_get_cells(num_cells)
            elif config.environment == "vscode":
                idx, cells_above = vscode_get_cells(num_cells)
            else:
                warnings.warn("Only nbclassic and vscode environments can access previous cells")
            prev_messages = cells_to_messages(cells_above)

    if is_config:
        handle_config(cell, config)
        return

    if idx is None:
        if config.environment == "nbclassic":
            idx, _ = nbclassic_get_cells(0)
        elif config.environment == "vscode":
            idx, _ = vscode_get_cells(0)

    # If it's a folllow-up message, we should first run the current cell and collect its outputs,
    # then pass the outputs to the chat as user messages for further generation.

    disp_capturer = PassthroughCapturer()
    if is_reply:
        disp_capturer.start()

        sys.stdout = stdout = TeeIO(sys.stdout)
        sys.stderr = stderr = TeeIO(sys.stderr)

        traceback = SaveTraceback(ip._showtraceback)
        traceback.register()

        # This runs the current cell, and we should capture all its outputs.
        res = ip.run_cell(cell)

        traceback.unregister()
        disp_capturer.stop()
        sys.stdout = stdout._stream
        sys.stderr = stderr._stream

        new_cell = make_cell_from_captured(source="%%fr " + line + "\n" + cell,
                                           res=res,
                                           stdout=stdout.getvalue(),
                                           stderr=stderr.getvalue(),
                                           displays=disp_capturer.get_outputs(),
                                           tracebacks=traceback.tracebacks)
        prev_messages += cells_to_messages([new_cell])

    else:
        prev_messages.append(nict(mk_msg(cell, role="user")))

    prev_messages = merge_messages(prev_messages)
    # Assistant message can't be the first in the conversation, nor can it be the last.
    if prev_messages:
        if prev_messages[0].role == "assistant":
            prev_messages.insert(0, nict(mk_msg("(Beginning of the conversation truncated)", role="user")))
        if prev_messages[-1].role == "assistant":
            prev_messages.append(nict(mk_msg("Continue", role="user")))

    # md_reply
    code_idx, md_display = chat.go(idx, num_cells, prev_messages)

    # If it's a reply cell, comment out the magic. Otherwise, comment out the
    # whole cell, as it's likely plain text, and not valid Python.
    # This makes it easy to re-run the ai-generated notebook manually without triggering the ai.
    cell_text = f"%%fr {line}\n{cell.strip()}".splitlines()
    if is_reply:
        cell_text[0] = "# " + cell_text[0]
    elif config.comment_after_run:
        cell_text = ["# " + l for l in cell_text]

    update_code_self("\n".join(cell_text))

    # This schedules the execution of the next cell.
    if (config.autorun
        and idx is not None
        and code_idx is not None
        and code_idx != idx):
        nbclassic_execute_cell(code_idx)

    # We might have created some invisible JS display outputs. Unfortunaely I don't think there is a way
    # to clear them without removing everying.
    clear_output()

    if is_reply:
        # Re-display the outputs.
        if stdout.getvalue(): print(stdout.getvalue())
        if stderr.getvalue(): print(stderr.getvalue(), file=sys.stderr)

        for d in disp_capturer.raw_outputs:
            ip.display_pub.publish(**d)
        for tb in traceback.tracebacks:
            ip._showtraceback(tb.ename, tb.evalue, tb.traceback)

    if md_display: display(md_display)
    if is_reply: return res.result



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_capture_trace.ipynb.

# %% auto 0
__all__ = ['SaveTraceback']

# %% ../nbs/03_capture_trace.ipynb 2
import re
from IPython import get_ipython
from .utils import nict

# %% ../nbs/03_capture_trace.ipynb 3
class SaveTraceback:
    def __init__(self, saved):
        self._saved_showtraceback = saved
        self.tracebacks = []

    def _showtraceback(self, etype, value, stb):
        self.tracebacks.append( nict(
            ename=etype,
            evalue=", ".join(value.args),
            traceback= stb # ["\n".join([ strip_junk(s) for s in stb]) if isinstance(stb, list) else stb]
        ))
        self._saved_showtraceback(etype, value, stb)

    def register(self):
        get_ipython()._showtraceback = self._showtraceback

    def unregister(self):
        get_ipython()._showtraceback = self._saved_showtraceback




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_capture_io.ipynb.

# %% auto 0
__all__ = ['TeeIO']

# %% ../nbs/03_capture_io.ipynb 2
from io import StringIO

# %% ../nbs/03_capture_io.ipynb 3
class TeeIO:
    """
    OutStream that and also passes it to the original stream.
    """
    def __init__(self, stream):
        self._io = StringIO()
        self._stream = stream
        self._call_log = []

    def write(self, s):
        self._io.write(s)
        return self._stream.write(s)

    def getvalue(self):
        return self._io.getvalue()

    # ipykernel's OutStream class defined some other things, pass them to the parent.
    def __getattr__(self, name):
        if name in ['_stream', '_io', '_call_log']:
            # This will call object.__getattr__ on self, returning the real self.value
            return super().__getattr__(name)
        self._call_log.append(("getattr", name))
        return getattr(self._stream, name)

    def __setattr__(self, name, value):
        if name in ['_stream', '_io', '_call_log']:
            super().__setattr__(name, value)
        else:
            self._call_log.append(("setattr", name, value))
            setattr(self._stdout, name, value)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_config.ipynb.

# %% auto 0
__all__ = ['strip_comments', 'handle_config']

# %% ../nbs/03_config.ipynb 1
# from friendlly.utils import nict
from .notebook import detect_environment, update_code_self
import warnings

# %% ../nbs/03_config.ipynb 2
def strip_comments(source: str):
    """Strips Python # comments"""
    lines = source.splitlines()
    return "\n".join([ l.split("#")[0] for l in lines ]).strip()

def handle_config(cell, global_config):

    cell = strip_comments(cell)
    # If the cell was empty, show the config
    if not cell:
        def value_to_str(v):
            if isinstance(v, str):
                if "\n" in v:
                    return f'"""{v}"""'
                else:
                    return f"'{str(v)}'"
            return str(v)

        cell = "\n".join([f"{k}={value_to_str(v)}" for k, v in global_config.items()])
        update_code_self("%%fr config\n" + cell)
    else:
        new_config = {}
        # Code, globals, locals. We want the locals
        exec(cell, {}, new_config)
        global_config.update(new_config)

        if global_config.autorun and global_config.environment != "nbclassic":
            warnings.warn("Autorun is only supported in nbclassic environment at the moment.")

        if global_config.md_cells and global_config.environment != "nbclassic":
            warnings.warn("Inserting Markdown cells is only supported in nbclassic environment at the moment.")
        if global_config.api_key:
            warnings.warn("FIY It's safer to set ANTHROPIC_API_KEY in the environment or even better .env file and use dotenv package to load it.")



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_magic_line.ipynb.

# %% auto 0
__all__ = ['parse_cell', 'fr_line']

# %% ../nbs/01_magic_line.ipynb 5
import json
from typing import List, Dict, Tuple
import textwrap
import claudette
from IPython import get_ipython
from IPython.display import display, clear_output, Markdown, Javascript

# %% ../nbs/01_magic_line.ipynb 6
# A single cell can contain multiple messages.
# A message is either a user message (starts with %fr) or a bot message (starts with #).
# Both can be multiline.
def parse_cell(
    cell: str # The raw body of the cell
) -> Tuple[List[Dict[str, str]], int]:
    """
    A single cell can contain multiple messages.
    A message is either a user message (starts with %fr) or a bot message (starts with #).
    Both can be multiline.

    Returns: a list of messages (with 'role' and 'content') and the number of %fr magics in the cell
    """
    parsed_lines = []
    num_magic = 0
    for line in cell.split('\n'):
        if line.startswith('%fr'):
            message = {'role': 'user', 'content': line[3:].strip()}
            num_magic += 1
        elif line.strip().startswith('#'):
            message = {'role': 'assistant', 'content': line[1:].strip()}
        else: continue

        if not parsed_lines or parsed_lines[-1]['role'] != message['role']:
            parsed_lines.append(message)
        else:
            parsed_lines[-1]['content'] += ("\n" + message['content'])

    return parsed_lines, num_magic

# %% ../nbs/01_magic_line.ipynb 7
models = [
    'claude-3-opus-20240229',
    'claude-3-5-sonnet-20240620',
    'claude-3-haiku-20240307',
]
chat_client = claudette.Client(model=models[1])

magic_count = 0
messages = []

# %% ../nbs/01_magic_line.ipynb 8
def fr_line(line: str):
    """The magic function for the %fr magic command."""
    global magic_count, messages
    ip = get_ipython()
    # raw_cell = ip.history_manager.input_hist_raw[-1]
    raw_cell = ip.get_parent()["content"]["code"]

    # The cell might have multiple %lm magics, but we only want to process the last one.
    # Presumably, the previous ones would have been processed already.
    if magic_count <= 0:
        messages, magic_count = parse_cell(raw_cell)

    # This is the last %lm magic invocation of the cell.
    # But we ignore cells that don't have a user message as the last message.
    if magic_count == 1 and len(messages) > 0 and messages[-1]['role'] == 'user' and messages[-1]['content'].strip():
        reply = ""
        display_id = display(Markdown("🚀..."), display_id=True)
        try:
            r = chat_client([m['content'] for m in messages], stream=True)
            for token in r:
                reply += token
                display_id.update(Markdown(reply))

            if reply:
                reply = textwrap.fill(text=reply, width=100, initial_indent="# ", subsequent_indent="# ")
                raw_cell += f"\n{reply}\n\n%fr "
                ip.set_next_input(raw_cell, replace=True)

            clear_output()

        except BaseException as e:
            display_id.update(Markdown(f"🚫 {repr(e)}"))


    magic_count -= 1



__version__ = "0.2.5"

from .magic_cell import fr_cell
from .magic_line import fr_line
from .notebook import nbclassic_patch_kernel, detect_environment
from IPython.display import clear_output

def load_ipython_extension(ipython):
    if detect_environment() == "nbclassic":
        nbclassic_patch_kernel()
        clear_output()

    ipython.register_magic_function(fr_cell, "cell", magic_name="fr")
    ipython.register_magic_function(fr_line, "line", magic_name="fr")

def unload_ipython_extension(ipython):
    pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_notebook.ipynb.

# %% auto 0
__all__ = ['update_code_self', 'update_code_next', 'detect_environment', 'vscode_extract_path', 'vscode_get_cells',
           'nbclassic_patch_kernel', 'nbclassic_add_cell', 'nbclassic_update_cell', 'nbclassic_execute_cell',
           'nbclassic_render_cell', 'nbclassic_delete_cell', 'nbclassic_get_cells']

# %% ../nbs/05_notebook.ipynb 2
from IPython import get_ipython
from IPython.display import display, clear_output, Markdown, Javascript
import json
import os
import time
import psutil
from urllib.parse import urlparse
import warnings
from .utils import nict

# %% ../nbs/05_notebook.ipynb 4
def update_code_self(source:str):
    """
    Updates the current cell with the contents of the cell passed as argument.

    The update will take place after the cell has finished executing.
    Works with all front-ennds.
    """

    payload = dict(
        source="set_next_input",
        text=source,
        replace=True,
        clear_output=False
    )
    # We have to call payload_manager directly because ip.set_next_input does not have a clear_output parameter
    get_ipython().payload_manager.write_payload(payload, single=False)


# %% ../nbs/05_notebook.ipynb 5
def update_code_next(source:str):
    """
    Updates the next cell with the code passed as argument.

    The update will take place after the cell has finished executing.
    Works with all front-ennds.
    """
    payload = dict(
        source="set_next_input",
        text=source,
        replace=False,
        clear_output=False
    )
    # We have to call payload_manager directly because ip.set_next_input does not have a clear_output parameter
    get_ipython().payload_manager.write_payload(payload, single=False)

# %% ../nbs/05_notebook.ipynb 6
def detect_environment():
    """
    Detects if we are running in vscode, Jupyter nbclassic, or Jupyter notebook 7 / Jupyter lab.
    """

    parent_cmdline = psutil.Process(os.getppid()).cmdline()

    for l in parent_cmdline:
        # Vscode might have a different name (e,g, cursor), but it should have "vscode" somewhere in the command line.
        if "vscode" in l.lower():
            return "vscode"

        # jupyter-nbclassic is the modern name of the "old" Jupyter notebook, as it was in Jupyter 6.
        if "jupyter-nbclassic" in l:
            return "nbclassic"

        # jupyter-notebook is probably Jupyter notebook 7, which is a re-skin of Jupyter lab
        if "jupyter-lab" in l or "jupyter-notebook" in l:
            return "jupyterlab"

    warnings.warn("Could not detect environment. Functionality might be limited.")
    return "unknown"

# %% ../nbs/05_notebook.ipynb 7
def inject_js(js:str):
    """Inject javascript into the notebook

    Args:
        js (str): The javascript code to be injected.

    After the injection, it will be overwrittent with an empty string to avoid running it again on notebook load.
    """

    display_handle = display(Javascript(js), display_id=True)
    display_handle.update(Javascript(""))

# %% ../nbs/05_notebook.ipynb 9
def vscode_extract_path():
    """
    Extracts the filename from the parent_header of the current notebook.
    """
    cellid = get_ipython().parent_header.get("metadata", {}).get("cellId", '')
    url = urlparse(cellid)
    return url.path

def vscode_get_cells(num_cells):
    path = vscode_extract_path()

    def get_last_modified(path):
        try: return os.path.getmtime(path)
        except BaseException as e:
            warnings.warn(f"{e}: Could not get mtime for {path}")
            return 0

    if path:
        exec_cnt = get_ipython().execution_count
        ts = time.time()

        display(Javascript("")) # Empty js to kick off autosave.
        last_modified = get_last_modified(path)

        # Wait for the file to be saved.
        for i in range(50):
            last_modified = get_last_modified(path)
            if last_modified > ts:
                break
            time.sleep(0.1)
        else:
            warnings.warn("Make sure autosave is set to afterDelay in vscode settings, and the delay is less than a second!")

        for i in range(50):
            try: # In case the file is being written to and is not parseable.
                with open(path) as f:
                    data = json.load(f)
                    cells = data.get("cells", [])
                    for idx, cell in enumerate(cells):
                        if cell.get("execution_count") == exec_cnt:
                            cells = [ nict(c) for c in cells[(max(0, idx-num_cells)):idx] ]
                            # source is saved as an array of strings in ipynb.
                            for c in cells: c.source = "\n".join(c.source)
                            return idx, cells
            except: pass
            time.sleep(0.1)

        warnings.warn(f"Could not find the cell data in {path} . The assistant won't be able to see previous cells.")
    else:
        warnings.warn("VSCode did not send a cellId in the parent_header. The assistant won't be able to see previous cells.")

    return None, []

# %% ../nbs/05_notebook.ipynb 13
def nbclassic_patch_kernel():
    """
    Overrides Codecell.execute to add cell_index cell_id,
    and possibly cells_above to the extras passed to the kernel.

    The execute() is based on the one in Jupyter nbclassic with the features added.
    """

    payload = """
        console.log("patching nbclassic execute function...")
        Jupyter.CodeCell.prototype.execute = function (stop_on_error) {
            if (!this.kernel) {
                console.log(i18n.msg._("Can't execute cell since kernel is not set."));
                return;
            }

            if (stop_on_error === undefined) {
                if (this.metadata !== undefined &&
                        this.metadata.tags !== undefined) {
                    if (this.metadata.tags.indexOf('raises-exception') !== -1) {
                        stop_on_error = false;
                    } else {
                        stop_on_error = true;
                    }
                } else {
                stop_on_error = true;
                }
            }

            this.clear_output(false, true);
            var old_msg_id = this.last_msg_id;
            if (old_msg_id) {
                this.kernel.clear_callbacks_for_msg(old_msg_id);
                delete Jupyter.CodeCell.msg_cells[old_msg_id];
                this.last_msg_id = null;
            }
            if (this.get_text().trim().length === 0) {
                // nothing to do
                this.set_input_prompt(null);
                return;
            }
            this.set_input_prompt('*');
            this.element.addClass("running");
            var callbacks = this.get_callbacks();


            const cell_index = Jupyter.notebook.find_cell_index(this)

            let extras = {
                cell_index : cell_index,
                cell_id: this.id
            }
            let text = this.get_text().trim()
            let firstLine = text.split('\\n')[0];

            // Parse the magic command
            if (firstLine.startsWith("%%fr")) {
                // Separate args by spaces or tabs
                let parts = firstLine.split(/\\s+|\\t+/);
                let magic = parts[0];
                if (parts.length > 1) {
                    let magic_args = parts.slice(1);
                    let plusNArg = magic_args.find(arg => arg.startsWith('+') && !isNaN(parseInt(arg.slice(1))));
                    if (plusNArg) {
                        let n = parseInt(plusNArg.slice(1));
                        let start_pos = Math.max(0, cell_index - n);
                        cells = Jupyter.notebook.get_cells().slice(start_pos, cell_index);

                        extras = {
                            cells_above: cells,
                            ...extras
                        }
                    }
                }
            }

            this.last_msg_id = this.kernel.execute(
                this.get_text(),
                callbacks,
                {silent: false, store_history: true, stop_on_error : stop_on_error, ...extras });
            Jupyter.CodeCell.msg_cells[this.last_msg_id] = this;
            this.render();
            this.events.trigger('execute.CodeCell', {cell: this});
            var that = this;
            function handleFinished(evt, data) {
                if (that.kernel.id === data.kernel.id && that.last_msg_id === data.msg_id) {
                        that.events.trigger('finished_execute.CodeCell', {cell: that});
                    that.events.off('finished_iopub.Kernel', handleFinished);
                }
            }
            this.events.on('finished_iopub.Kernel', handleFinished);
        };
        Jupyter.notebook.events.trigger('set_dirty.Notebook', {value: true});
        Jupyter.notebook._fully_patched = true;
        console.log("Done.")
    """
    inject_js(payload)

# %% ../nbs/05_notebook.ipynb 14
def nbclassic_add_cell(
        idx:int = None,         # Index of the cell to add. If none, add the cell under the selected one.
        cell_type:str = "code"  # Type of cell to add. Can be "code", "markdown", "raw"
    ):
    """
    Add a new notebook cell to an nbclassic notebook.
    Uses nbclassic-specific JS injection.
    """
    if not idx:
        index_payload = "let index = Jupyter.notebook.get_selected_index()+1;"
    else:
        index_payload = f"let index = {idx}"

    payload = f"""
        {index_payload}

        Jupyter.notebook.insert_cell_at_index("{cell_type}", index)
        let cell = Jupyter.notebook.get_cell(index);
        cell.events.trigger('set_dirty.Notebook', {{value: true}});
    """

    inject_js(payload)

# %% ../nbs/05_notebook.ipynb 15
def nbclassic_update_cell(
    idx:int, # Index of the cell to update. None to update the current cell
    text:str, # Text to set in the cell
    flush:bool = True # Notify Jupyter that the cell has been updated.
    ):
    """
    Update the text of a cell in an nbclassic notebook.
    Uses nbclassic-specific JS injection.
    """

    def escape_for_js(text):
        # Use json.dumps to escape the string for JavaScript
        escaped = json.dumps(text)
        # Remove the surrounding quotes added by json.dumps
        escaped = escaped[1:-1]
        # Escape backticks and ${} sequences
        return escaped.replace('`', '\\`').replace('${', '\\${')

    payload = f"""
        let cell = Jupyter.notebook.get_cell({idx})
        cell.set_text(`{escape_for_js(text)}`)
    """
    if flush:
         payload = payload + "\nJupyter.notebook.events.trigger('set_dirty.Notebook', {value: true});"
    inject_js(payload)

# %% ../nbs/05_notebook.ipynb 16
def nbclassic_execute_cell(
        idx:int # Index of the cell to execute. They start at 0
    ):
    """
    Execute a cell in an nbclassic notebook.
    Uses nbclassic-specific JS injection.
    """

    payload = f"""
        console.log("execute_cell", {idx});
        Jupyter.notebook.events.trigger('set_dirty.Notebook', {{value: true}});
        let cell = Jupyter.notebook.get_cell({idx})
        cell.execute()
    """
    # tt = display(f"About to run the cell {idx}...", display_id=True)
    inject_js(payload)

# %% ../nbs/05_notebook.ipynb 17
def nbclassic_render_cell(idx:int): # Cell to render.
    """
    Re-render a cell in an nbclassic notebook.
    Uses nbclassic-specific JS injection.
    """

    payload = f"""
        let cell = Jupyter.notebook.get_cell({idx})
        cell.unrender()
        Jupyter.notebook.events.trigger('set_dirty.Notebook', {{value: true}});

        cell.render()
    """

    inject_js(payload)

# %% ../nbs/05_notebook.ipynb 18
def nbclassic_delete_cell(idx:int): # Cell to delete.
    """
    Delete a cell in an nbclassic notebook.
    Uses nbclassic-specific JS injection.
    """

    payload = f"""
        console.log("deleting cell", {idx});
        Jupyter.notebook.delete_cell({idx});
        Jupyter.notebook.events.trigger('set_dirty.Notebook', {{value: true}});
    """

    inject_js(payload)

# %% ../nbs/05_notebook.ipynb 19
def nbclassic_get_cells(num_cells:int):
    """
    Get the cells in an nbclassic notebook.
    The notebook should have been patched with nbclassic_patch_kernel.
    """

    header = nict(get_ipython().parent_header)

    cell_index = header.get("content", {}).get("cell_index", None)
    if cell_index is None:
        warnings.warn("Jupyter did not send the cell index. Has it been patched with nbclassic_patch_kernel?")

    expected_num_cells = min(cell_index, num_cells)
    cells = header.get("content", {}).get("cells_above", [])
    if len(cells) < expected_num_cells:
        warnings.warn(f"Expected {expected_num_cells} cells, but got {len(cells)}")

    return cell_index, cells



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_display.ipynb.

# %% auto 0
__all__ = ['PassthroughCapturer']

# %% ../nbs/02_display.ipynb 2
from IPython.core.displaypub import DisplayPublisher
from IPython.core.display_functions import _new_id
from IPython.display import display, clear_output, Javascript
from IPython import get_ipython
from typing import List
from PIL import Image
from collections import defaultdict
import warnings
import sys

# %% ../nbs/02_display.ipynb 3
class PassthroughCapturer(DisplayPublisher):
    """A DisplayPublisher that stores AND displays!"""

    raw_outputs: List

    def __init__(self):
        self._publisher = None
        self.raw_outputs = []

    def publish(
        self, data, metadata=None, source=None, *, transient=None, update=False
    ):
        self.raw_outputs.append(
            {
                "data": data,
                "metadata": metadata,
                "transient": transient,
                "update": update,
            }
        )
        self._publisher.publish(data, metadata=metadata, transient=transient, update=update)

    def clear_output(self, wait=False):
        self.raw_outputs = []
        self._publisher.clear_output(wait)

    def start(self):
        ip = get_ipython()
        self._publisher = ip.display_pub
        ip.display_pub = self

        # print("starting passthrough: ", id(self._publisher), id(self))

    # Note: stop must be called from the cell that started the
    # passthrough, or from the associated post-cell callback
    def stop(self):
        ip = get_ipython()
        assert self._publisher
        ip.display_pub = self._publisher

    def get_outputs(self, with_js=False):
        coalesced_outputs = []
        display_id_map = defaultdict(list)

        outputs = self.raw_outputs
        if not with_js:
            outputs = [o for o in outputs if 'application/javascript' not in o['data']]

        # dicts are ordered in Python 3.7+
        for output in outputs:
            display_id = output['transient']['display_id'] if output['transient'] else _new_id()
            if not output['update']:
                display_id_map[display_id] += [output["data"]]
            else:
                display_id_map[display_id][-1] = output["data"]

        for display_id, output_list in display_id_map.items():
            coalesced_outputs += output_list

        return coalesced_outputs



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/friendlly',
                'doc_host': 'https://xl0.github.io',
                'git_url': 'https://github.com/xl0/friendlly',
                'lib_path': 'friendlly'},
  'syms': { 'friendlly.capture_display': { 'friendlly.capture_display.PassthroughCapturer': ( 'capture_display.html#passthroughcapturer',
                                                                                              'friendlly/capture_display.py'),
                                           'friendlly.capture_display.PassthroughCapturer.__init__': ( 'capture_display.html#passthroughcapturer.__init__',
                                                                                                       'friendlly/capture_display.py'),
                                           'friendlly.capture_display.PassthroughCapturer.clear_output': ( 'capture_display.html#passthroughcapturer.clear_output',
                                                                                                           'friendlly/capture_display.py'),
                                           'friendlly.capture_display.PassthroughCapturer.get_outputs': ( 'capture_display.html#passthroughcapturer.get_outputs',
                                                                                                          'friendlly/capture_display.py'),
                                           'friendlly.capture_display.PassthroughCapturer.publish': ( 'capture_display.html#passthroughcapturer.publish',
                                                                                                      'friendlly/capture_display.py'),
                                           'friendlly.capture_display.PassthroughCapturer.start': ( 'capture_display.html#passthroughcapturer.start',
                                                                                                    'friendlly/capture_display.py'),
                                           'friendlly.capture_display.PassthroughCapturer.stop': ( 'capture_display.html#passthroughcapturer.stop',
                                                                                                   'friendlly/capture_display.py')},
            'friendlly.capture_io': { 'friendlly.capture_io.TeeIO': ('capture_io.html#teeio', 'friendlly/capture_io.py'),
                                      'friendlly.capture_io.TeeIO.__getattr__': ( 'capture_io.html#teeio.__getattr__',
                                                                                  'friendlly/capture_io.py'),
                                      'friendlly.capture_io.TeeIO.__init__': ('capture_io.html#teeio.__init__', 'friendlly/capture_io.py'),
                                      'friendlly.capture_io.TeeIO.__setattr__': ( 'capture_io.html#teeio.__setattr__',
                                                                                  'friendlly/capture_io.py'),
                                      'friendlly.capture_io.TeeIO.getvalue': ('capture_io.html#teeio.getvalue', 'friendlly/capture_io.py'),
                                      'friendlly.capture_io.TeeIO.write': ('capture_io.html#teeio.write', 'friendlly/capture_io.py')},
            'friendlly.capture_trace': { 'friendlly.capture_trace.SaveTraceback': ( 'capture_trace.html#savetraceback',
                                                                                    'friendlly/capture_trace.py'),
                                         'friendlly.capture_trace.SaveTraceback.__init__': ( 'capture_trace.html#savetraceback.__init__',
                                                                                             'friendlly/capture_trace.py'),
                                         'friendlly.capture_trace.SaveTraceback._showtraceback': ( 'capture_trace.html#savetraceback._showtraceback',
                                                                                                   'friendlly/capture_trace.py'),
                                         'friendlly.capture_trace.SaveTraceback.register': ( 'capture_trace.html#savetraceback.register',
                                                                                             'friendlly/capture_trace.py'),
                                         'friendlly.capture_trace.SaveTraceback.unregister': ( 'capture_trace.html#savetraceback.unregister',
                                                                                               'friendlly/capture_trace.py')},
            'friendlly.config': { 'friendlly.config.handle_config': ('config.html#handle_config', 'friendlly/config.py'),
                                  'friendlly.config.strip_comments': ('config.html#strip_comments', 'friendlly/config.py')},
            'friendlly.display': { 'friendlly.display.PassthroughCapturer': ('display.html#passthroughcapturer', 'friendlly/display.py'),
                                   'friendlly.display.PassthroughCapturer.__init__': ( 'display.html#passthroughcapturer.__init__',
                                                                                       'friendlly/display.py'),
                                   'friendlly.display.PassthroughCapturer.clear_output': ( 'display.html#passthroughcapturer.clear_output',
                                                                                           'friendlly/display.py'),
                                   'friendlly.display.PassthroughCapturer.get_outputs': ( 'display.html#passthroughcapturer.get_outputs',
                                                                                          'friendlly/display.py'),
                                   'friendlly.display.PassthroughCapturer.publish': ( 'display.html#passthroughcapturer.publish',
                                                                                      'friendlly/display.py'),
                                   'friendlly.display.PassthroughCapturer.start': ( 'display.html#passthroughcapturer.start',
                                                                                    'friendlly/display.py'),
                                   'friendlly.display.PassthroughCapturer.stop': ( 'display.html#passthroughcapturer.stop',
                                                                                   'friendlly/display.py')},
            'friendlly.magic_cell': { 'friendlly.magic_cell.CellChat': ('magic_cell.html#cellchat', 'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.CellChat.__init__': ( 'magic_cell.html#cellchat.__init__',
                                                                                  'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.CellChat.delete_markdown': ( 'magic_cell.html#cellchat.delete_markdown',
                                                                                         'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.CellChat.go': ('magic_cell.html#cellchat.go', 'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.CellChat.update_code': ( 'magic_cell.html#cellchat.update_code',
                                                                                     'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.CellChat.update_code_final': ( 'magic_cell.html#cellchat.update_code_final',
                                                                                           'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.CellChat.update_markdown': ( 'magic_cell.html#cellchat.update_markdown',
                                                                                         'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.NBClassicChat': ('magic_cell.html#nbclassicchat', 'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.NBClassicChat.__init__': ( 'magic_cell.html#nbclassicchat.__init__',
                                                                                       'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.NBClassicChat.delete_markdown': ( 'magic_cell.html#nbclassicchat.delete_markdown',
                                                                                              'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.NBClassicChat.update_code': ( 'magic_cell.html#nbclassicchat.update_code',
                                                                                          'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.NBClassicChat.update_code_final': ( 'magic_cell.html#nbclassicchat.update_code_final',
                                                                                                'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.NBClassicChat.update_markdown': ( 'magic_cell.html#nbclassicchat.update_markdown',
                                                                                              'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.NoJSChat': ('magic_cell.html#nojschat', 'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.NoJSChat.__init__': ( 'magic_cell.html#nojschat.__init__',
                                                                                  'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.NoJSChat.delete_markdown': ( 'magic_cell.html#nojschat.delete_markdown',
                                                                                         'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.NoJSChat.update_code': ( 'magic_cell.html#nojschat.update_code',
                                                                                     'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.NoJSChat.update_code_final': ( 'magic_cell.html#nojschat.update_code_final',
                                                                                           'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.NoJSChat.update_markdown': ( 'magic_cell.html#nojschat.update_markdown',
                                                                                         'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.cells_to_messages': ( 'magic_cell.html#cells_to_messages',
                                                                                  'friendlly/magic_cell.py'),
                                      'friendlly.magic_cell.fr_cell': ('magic_cell.html#fr_cell', 'friendlly/magic_cell.py')},
            'friendlly.magic_line': { 'friendlly.magic_line.fr_line': ('magic_line.html#fr_line', 'friendlly/magic_line.py'),
                                      'friendlly.magic_line.parse_cell': ('magic_line.html#parse_cell', 'friendlly/magic_line.py')},
            'friendlly.notebook': { 'friendlly.notebook.detect_environment': ('notebook.html#detect_environment', 'friendlly/notebook.py'),
                                    'friendlly.notebook.inject_js': ('notebook.html#inject_js', 'friendlly/notebook.py'),
                                    'friendlly.notebook.nbclassic_add_cell': ('notebook.html#nbclassic_add_cell', 'friendlly/notebook.py'),
                                    'friendlly.notebook.nbclassic_delete_cell': ( 'notebook.html#nbclassic_delete_cell',
                                                                                  'friendlly/notebook.py'),
                                    'friendlly.notebook.nbclassic_execute_cell': ( 'notebook.html#nbclassic_execute_cell',
                                                                                   'friendlly/notebook.py'),
                                    'friendlly.notebook.nbclassic_get_cells': ( 'notebook.html#nbclassic_get_cells',
                                                                                'friendlly/notebook.py'),
                                    'friendlly.notebook.nbclassic_patch_kernel': ( 'notebook.html#nbclassic_patch_kernel',
                                                                                   'friendlly/notebook.py'),
                                    'friendlly.notebook.nbclassic_render_cell': ( 'notebook.html#nbclassic_render_cell',
                                                                                  'friendlly/notebook.py'),
                                    'friendlly.notebook.nbclassic_update_cell': ( 'notebook.html#nbclassic_update_cell',
                                                                                  'friendlly/notebook.py'),
                                    'friendlly.notebook.update_code_next': ('notebook.html#update_code_next', 'friendlly/notebook.py'),
                                    'friendlly.notebook.update_code_self': ('notebook.html#update_code_self', 'friendlly/notebook.py'),
                                    'friendlly.notebook.vscode_extract_path': ( 'notebook.html#vscode_extract_path',
                                                                                'friendlly/notebook.py'),
                                    'friendlly.notebook.vscode_get_cells': ('notebook.html#vscode_get_cells', 'friendlly/notebook.py')},
            'friendlly.utils': { 'friendlly.utils.GenericEncoder': ('utils.html#genericencoder', 'friendlly/utils.py'),
                                 'friendlly.utils.GenericEncoder.default': ('utils.html#genericencoder.default', 'friendlly/utils.py'),
                                 'friendlly.utils.cell_to_blocks': ('utils.html#cell_to_blocks', 'friendlly/utils.py'),
                                 'friendlly.utils.display_data_to_blocks': ('utils.html#display_data_to_blocks', 'friendlly/utils.py'),
                                 'friendlly.utils.make_cell_from_captured': ('utils.html#make_cell_from_captured', 'friendlly/utils.py'),
                                 'friendlly.utils.merge_messages': ('utils.html#merge_messages', 'friendlly/utils.py'),
                                 'friendlly.utils.nict': ('utils.html#nict', 'friendlly/utils.py'),
                                 'friendlly.utils.nict.__delattr__': ('utils.html#nict.__delattr__', 'friendlly/utils.py'),
                                 'friendlly.utils.nict.__dir__': ('utils.html#nict.__dir__', 'friendlly/utils.py'),
                                 'friendlly.utils.nict.__getattr__': ('utils.html#nict.__getattr__', 'friendlly/utils.py'),
                                 'friendlly.utils.nict.__init__': ('utils.html#nict.__init__', 'friendlly/utils.py'),
                                 'friendlly.utils.nict.__setattr__': ('utils.html#nict.__setattr__', 'friendlly/utils.py'),
                                 'friendlly.utils.nict._to_dict': ('utils.html#nict._to_dict', 'friendlly/utils.py'),
                                 'friendlly.utils.parse_cell_line': ('utils.html#parse_cell_line', 'friendlly/utils.py'),
                                 'friendlly.utils.safe_int': ('utils.html#safe_int', 'friendlly/utils.py'),
                                 'friendlly.utils.strip_traceback_junk': ('utils.html#strip_traceback_junk', 'friendlly/utils.py'),
                                 'friendlly.utils.svg_to_png': ('utils.html#svg_to_png', 'friendlly/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/06_utils.ipynb.

# %% auto 0
__all__ = ['nict', 'GenericEncoder', 'safe_int', 'strip_traceback_junk', 'parse_cell_line', 'svg_to_png',
           'display_data_to_blocks', 'cell_to_blocks', 'make_cell_from_captured', 'merge_messages']

# %% ../nbs/06_utils.ipynb 1
import pprint

# %% ../nbs/06_utils.ipynb 2
class nict(dict):
    def __init__(self, d=dict(), /, **kwargs):
        if d is None: d = {}
        else:
            assert isinstance(d, dict), f'expected dict, got {type(d)}'
            d = d.copy()

        d.update(kwargs)
        for k,v  in d.items():
            if isinstance(v, dict):
                d[k] = nict(**v)
            elif isinstance(v, list):
                d[k] = [nict(**item) if isinstance(item, dict) else item for item in v]
            elif isinstance(v, tuple):
                d[k] = tuple(nict(**item) if isinstance(item, dict) else item for item in v)
            else:
                d[k] = v

        super().__init__(**d)

    def __getattr__(self, key):
        return self[key] if key in self else super().__getattribute__(key)

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        del self[key]

    def __dir__(self):
        return list(self.keys()) + dir(super())

    def _to_dict(self):
        return {k: v._to_dict() if isinstance(v, nict) else v for k, v in self.items()}

# %% ../nbs/06_utils.ipynb 13
import json
from datetime import datetime

# %% ../nbs/06_utils.ipynb 14
class GenericEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        try:
            return obj.__dict__
        except AttributeError:
            try:
                return str(obj)
            except:
                return f"<{type(obj).__name__}>"

# %% ../nbs/06_utils.ipynb 15
def safe_int(s, default=0):
    try:
        return int(s)
    except ValueError:
        return default

# %% ../nbs/06_utils.ipynb 16
import re

# %% ../nbs/06_utils.ipynb 17
def strip_traceback_junk(text):
    """Strips colors and other junk from a traceback"""

    # This line just easts up tokens. We don't need it.
    text = text.replace("---------------------------------------------------------------------------", "")
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return ansi_escape.sub('', text)

# %% ../nbs/06_utils.ipynb 19
from claudette import text_msg, img_msg
import base64

# %% ../nbs/06_utils.ipynb 20
def parse_cell_line(line):
    num_cells, is_friendlly, is_reply, is_config = 0, False, False, False
    if line.lstrip('# \t').rstrip().startswith("%%fr"):
        is_friendlly = True
        args = line.split()
        for arg in args[1:]:
            if "+" in arg: num_cells = safe_int(arg[1:])
            if arg == "reply": is_reply = True
            if arg == "config" : is_config = True

    return num_cells, is_friendlly, is_reply, is_config

# %% ../nbs/06_utils.ipynb 21
import cairosvg
import xml.etree.ElementTree as ET

# %% ../nbs/06_utils.ipynb 22
def svg_to_png(svg_content, max_size):
    root = ET.fromstring(svg_content)

    def parse_dimension(value):
        if value.endswith('%'):
            return float(value[:-1]) / 100 * max_size
        return float(re.findall(r'[\d.]+', value)[0])

    width = parse_dimension(root.attrib.get('width', str(max_size)))
    height = parse_dimension(root.attrib.get('height', str(max_size)))

    viewBox = root.attrib.get('viewBox')
    if viewBox:
        _, _, vb_width, vb_height = map(float, viewBox.split())
        if width == max_size:
            width = vb_width
        if height == max_size:
            height = vb_height

    scale = min(max_size / width, max_size / height)
    new_width = int(width * scale)
    new_height = int(height * scale)

    # Render PNG
    png =  cairosvg.svg2png(bytestring=svg_content, output_width=new_width, output_height=new_height)
    return png

# %% ../nbs/06_utils.ipynb 24
def display_data_to_blocks(dsp):
    """
    Converts a display_data (a dict of mime types to data) to a list of claudette messages (text, image, with xml tags)
    """

    # The dsp for an svg might look like this:
    # {
    #     "text/plain": "<SVG object>",
    #     "image/svg+xml": "<svg>....</svg>",
    #     "image/png": "<base64 image>", # or binary string
    #     "image/jpeg": "<base64 image>", # or binary string
    # }
    # Or for markdown:
    # {
    #     "text/plain": "Markdown text",
    #     "text/markdown": "## Markdown text",
    # }

    plain_text = dsp.get("text/plain")

    # Map mime types to their xml tags
    text_types = list(filter(lambda x:x[1], [
        ("text-html", dsp.get("text/html")),
        ("latex", dsp.get("text/latex")),
        ("markdown", dsp.get("text/markdown")),
        ("json", dsp.get("application/json")),
        ("js", dsp.get("application/javascript"))
    ]))
    messages = [f"<{t[0]}>{t[1]}</{t[0]}>" for t in text_types]

    images = list(filter(lambda x:x, [ dsp.get("image/png"), dsp.get("image/jpeg") ]))
    # Svg is usually rendered by ipython as an image, but let's handle unrendered svg as well
    svg = dsp.get("image/svg+xml")
    if svg: images.append(svg_to_png(svg, 256))

    # Different text typed outputs come with plain text.
    # We only include the plain text if no other text type is present.
    if not len(messages):
        messages = [plain_text]

    claudette_blocks = [text_msg(m) for m in messages]
    if images:
        image = images[0] # Grab only one image, they are ordered as png, jpeg, rendered svg
        # If the image is a base64 string, convert it to binary
        if isinstance(image, str):
            image = base64.b64decode(image)
        claudette_blocks.append(img_msg(image))

    return claudette_blocks


def cell_to_blocks(cell):
    "Converts a cell and its outputs to a list of claudette messages (text, image, with xml tags)"

    assert "source" in cell
    if cell.cell_type != "code":
        assert not "outputs" in cell or not len(cell.outputs), "Got a non-code cell with outputs??"
        return False, False, text_msg(cell.source), []

    is_reply = False
    is_friendlly = False
    output_blks = []

    lines = cell.source.strip().splitlines()
    if lines:
        _, is_friendlly, is_reply, _ = parse_cell_line(lines[0])

        # Remove the cell magic
        if is_friendlly:
            lines = lines[1:]
        code_text = "\n".join(lines)

        outputs = cell.get("outputs", [])
        for o in outputs:
            if o.output_type == "stream":
                output_blks.append(text_msg(f"<{o.name}>{o.text}</{o.name}>"))
            if o.output_type in ("display_data", "execute_result"):
                # We produce a lot of empty JS outputs when running the chat in nbclassic mode,
                # and they don't get cleaned up in time. Ignore them.
                if "application/javascript" not in o.data or o.data["application/javascript"]:
                    output_blks += display_data_to_blocks(o.data)
            if o.output_type == "error":
                output_blks.append(text_msg("<error>" + strip_traceback_junk("\n".join(o.traceback)) + "</error>"))

        return is_friendlly, is_reply, code_text, output_blks
    return False, False, None, []

# %% ../nbs/06_utils.ipynb 25
from IPython import get_ipython

# %% ../nbs/06_utils.ipynb 26
def make_cell_from_captured(source:str, res, stdout:str, stderr:str, displays, tracebacks):
    ip = get_ipython()
    cell = nict()
    cell.metadata  = {}
    cell.cell_type = "code"
    cell.source    = source
    cell.outputs   = []
    if stdout: cell.outputs.append(nict(output_type="stream", name="stdout", text=stdout))
    if stderr: cell.outputs.append(nict(output_type="stream", name="stderr", text=stderr))
    for d in displays:
        cell.outputs.append(nict(output_type="display_data", data=d))
    if res.result:
        formatted, _ = ip.display_formatter.format(res.result)
        cell.outputs.append(nict(output_type="execute_result", data=formatted))
    for t in tracebacks:
        cell.outputs.append(nict(t, output_type="error"))

    return cell

# %% ../nbs/06_utils.ipynb 27
from typing import List
import claudette

# %% ../nbs/06_utils.ipynb 28
def merge_messages(msgs: List[nict]):
    """
    Takes a list of messages and merges them together, coalescing first by role,
    and then also coalescing text blocks together.
    """

    # first, merge by role
    role_messages = []
    prev_role = None

    for m in msgs:
        if prev_role == m.role: role_messages[-1].content += m.content
        else:                   role_messages.append(m)
        prev_role = m.role

    # Then, for assistant, also coalesce messages together
    for m in role_messages:
        merged_content = []
        for c in m.content:
            if c.type == "text" and merged_content and merged_content[-1].type == "text":
                merged_content[-1].text += ("\n" + c.text)
            else: merged_content.append(c)
        m.content = merged_content
    return role_messages

